#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
obspy.io.focmec - FOCMEC file format support for ObsPy
======================================================

This module provides read support for some FOCMEC file formats.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)
