# -*- coding: utf-8 -*-
"""
obspy.io.ascii - SLIST and TSPAIR support for ObsPy
===================================================

This module provides read and write support for the SLIST and TSPAIR ASCII
formats.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
